@@ -1,5 +1,3 @@
-# type: ignore
-# TODO(ameer): importing backend code in frontend is not making mypy happy.
 """Used by the backend and CLI.
 Placed in frontend/cli/anyscale package to be in the PYTHONPATH when
 autoscaler searches for the module.
@@ -14,24 +12,18 @@ import uuid
 
 import ray
 from ray.autoscaler.node_provider import NodeProvider
-from ray.autoscaler.util import fillout_defaults
 import yaml
 
-from anyscale.util import send_json_request, startup_log_name
-from backend.database.models import Session
+from anyscale.util import send_json_request
 from backend.server.services.gateway_commandrunner_service import (
     GatewaySSHCommandRunner,
 )
 from backend.server.services.gateway_router_service import GatewayID, GatewayRequest
-from backend.server.util import (
-    StdoutStderrRedirect,
-    temporary_db_session,
-)
 
 logger = logging.getLogger(__name__)
 
 
-class ServerNodeProvider(NodeProvider):
+class ServerNodeProvider(NodeProvider):  # type: ignore
     """Interface for getting and returning nodes from a Cloud gateway.
 
     Servers are namespaced by the `cluster_name` parameter,
@@ -43,7 +35,6 @@ class ServerNodeProvider(NodeProvider):
     Args:
         provider_config (dict): the provider section in the cluster_config.
         cluster_name (str): the cluster name.
-        cluster_config (dict): the cluster config. Passed only in testing mode.
     """
 
     def __init__(
