@@ -5,6 +5,7 @@ import datetime
 import json
 import logging
 import os
+from os.path import isfile
 from shlex import quote
 import shutil
 import subprocess
@@ -18,7 +19,6 @@ import uuid
 import click
 from jsonpatch import JsonPatch
 from ray.autoscaler.commands import (
-    create_or_update_cluster,
     exec_cluster,
     get_head_node_ip,
     get_worker_node_ips,
@@ -36,7 +36,7 @@ from anyscale.auth_proxy import app as auth_proxy_app
 from anyscale.autosync_heartbeat import managed_autosync_session
 from anyscale.cloudgateway import CloudGatewayRunner
 import anyscale.conf
-from anyscale.feature_flags import FLAG_KEY_USE_SNAPSHOT
+from anyscale.feature_flags import FLAG_KEY_USE_CLOUD, FLAG_KEY_USE_SNAPSHOT
 from anyscale.project import (
     ANYSCALE_AUTOSCALER_FILE,
     ANYSCALE_PROJECT_FILE,
@@ -64,6 +64,7 @@ from anyscale.util import (
     get_available_regions,
     get_endpoint,
     humanize_timestamp,
+    launch_gcp_cloud_setup,
     send_json_request,
     send_json_request_raw,
     slugify,
@@ -91,6 +92,7 @@ def get_or_create_snapshot(
             yes,
             description=description,
             tags=["anyscale:session_startup"],
+            run_upload=True,
         )
     else:
         snapshot_uuid = get_snapshot_uuid(project_definition.root, snapshot_uuid)
@@ -122,6 +124,20 @@ def get_project_session(project_id: int, session_name: Optional[str]) -> Any:
     return sessions[0]
 
 
+def get_user_cloud() -> Any:
+    response = send_json_request("/api/v2/clouds/", {})
+    clouds = response["results"]
+    if len(clouds) > 1:
+        raise click.ClickException(
+            "Multiple clouds: {}\n"
+            "Please specify the one you want to refer to.".format(
+                [cloud["name"] for cloud in clouds]
+            )
+        )
+
+    return clouds[0]
+
+
 def get_project_directory_name(project_id: int) -> str:
     resp = send_json_request("/api/v2/projects/", {})
     directory_name = ""
@@ -133,7 +149,7 @@ def get_project_directory_name(project_id: int) -> str:
     return directory_name
 
 
-def setup_ssh_for_head_node(session_id: int) -> Tuple[str, str]:
+def setup_ssh_for_head_node(session_id: int) -> Tuple[str, str, str]:
     resp = send_json_request("/api/v2/sessions/{}/ssh_key".format(session_id), {},)
     key_path = write_ssh_key(resp["result"]["key_name"], resp["result"]["private_key"])
 
@@ -143,7 +159,12 @@ def setup_ssh_for_head_node(session_id: int) -> Tuple[str, str]:
 
     resp = send_json_request("/api/v2/sessions/{}/head_ip".format(session_id), {})
     head_ip = resp["result"]["head_ip"]
-    return head_ip, key_path
+
+    config_resp = send_json_request(f"/api/v2/sessions/{session_id}/cluster_config", {})
+    config = json.loads(config_resp["result"]["config"])
+    config = fillout_defaults(config)
+    ssh_user = config.get("auth", {}).get("ssh_user")
+    return head_ip, key_path, ssh_user
 
 
 def update_file_mounts(
@@ -384,14 +405,23 @@ def drop_cloud_from_project(project_name: str, yes: bool) -> None:
 
 
 @cloud_cli.command(name="setup", help="Set up a cloud provider.")
-@click.option("--provider", help="Project to drop the cloud from.", required=True)
 @click.option(
-    "--region", help="Region to set up the credentials in.", default="us-west-2",
+    "--provider",
+    help="The cloud provider type.",
+    required=True,
+    prompt="Provider",
+    type=click.Choice(["aws", "gcp"], case_sensitive=False),
 )
-@click.option("--name", help="Name of the cloud.", required=True)
+@click.option(
+    "--region", help="Region to set up the credentials in.", default="us-west-2"
+)
+@click.option("--name", help="Name of the cloud.", required=True, prompt="Name")
 def setup_cloud(provider: str, region: str, name: str) -> None:
-    if provider == "AWS":
+    if provider == "aws":
         setup_aws(region, name)
+    elif provider == "gcp":
+        # TODO: interactive setup process through the CLI?
+        launch_gcp_cloud_setup()
 
 
 def setup_aws(region: str, name: str) -> None:
@@ -425,11 +455,27 @@ def setup_aws(region: str, name: str) -> None:
     print("Head over to the web UI to create new sessions in your AWS account!")
 
 
+def validate_cluster_configuration(cluster_config_file: str) -> None:
+    with open(cluster_config_file) as f:
+        cluster_config = yaml.safe_load(f)
+    try:
+        send_json_request(
+            "/api/v2/sessions/validate_cluster",
+            {"config": json.dumps(cluster_config)},
+            method="POST",
+        )
+    except Exception:
+        raise click.ClickException(
+            "The configuration file {} is not valid.".format(cluster_config_file)
+        )
+
+
 def register_project(project_definition: Any) -> None:
     project_id_path = os.path.join(ray_scripts.PROJECT_DIR, PROJECT_ID_BASENAME)
 
+    # validate_cluster_configuration(project_definition.cluster_yaml())
+
     project_name = project_definition.config["name"]
-    print("Creating project {}".format(project_name))
     description = project_definition.config.get("description", "")
 
     # Add a database entry for the new Project.
@@ -455,6 +501,9 @@ def register_project(project_definition: Any) -> None:
             False,
             description="Initial project snapshot",
             tags=["anyscale:initial"],
+            run_upload=(
+                os.environ.get("DEPLOY_ENVIRONMENT") != "test"
+            ),  # TODO(ilr) Always run on startup (until this is disabled)
         )
     except click.Abort as e:
         raise e
@@ -462,7 +511,6 @@ def register_project(project_definition: Any) -> None:
         # Creating a snapshot can fail if the project is not found or if some
         # files cannot be copied (e.g., due to permissions).
         raise click.ClickException(e)  # type: ignore
-
     # Print success message
     url = get_endpoint(f"/projects/{project_id}")
     print(f"Project {project_id} created. View at {url}")
@@ -501,7 +549,15 @@ def create_new_proj_def(
 
     # Create startup.yaml.
     if cluster_config_file:
-        shutil.copyfile(cluster_config_file, anyscale.project.ANYSCALE_AUTOSCALER_FILE)
+        # validate_cluster_configuration(cluster_config_file)
+        if not os.path.exists(
+            anyscale.project.ANYSCALE_AUTOSCALER_FILE
+        ) or not os.path.samefile(
+            cluster_config_file, anyscale.project.ANYSCALE_AUTOSCALER_FILE
+        ):
+            shutil.copyfile(
+                cluster_config_file, anyscale.project.ANYSCALE_AUTOSCALER_FILE
+            )
     else:
         if not os.path.exists(anyscale.project.ANYSCALE_AUTOSCALER_FILE):
             with open(anyscale.project.ANYSCALE_AUTOSCALER_FILE, "w") as f:
@@ -516,7 +572,7 @@ def create_new_proj_def(
 )
 @click.option("--name", help="Project name.", required=False)
 @click.option(
-    "--cluster-config-file",
+    "--config",
     help="Path to autoscaler yaml. Created by default.",
     type=click.Path(exists=True),
     required=False,
@@ -527,11 +583,9 @@ def create_new_proj_def(
     required=False,
 )
 @click.pass_context
+# flake8: noqa: C901
 def anyscale_init(
-    ctx: Any,
-    name: Optional[str],
-    cluster_config_file: Optional[str],
-    requirements: Optional[str],
+    ctx: Any, name: Optional[str], config: Optional[str], requirements: Optional[str],
 ) -> None:
     # Send an initial request to the server to make sure we are actually
     # registered. We only want to create the project if that is the case,
@@ -543,14 +597,11 @@ def anyscale_init(
         if os.path.exists(ray_scripts.PROJECT_DIR)
         else anyscale.project.ANYSCALE_PROJECT_FILE
     )
-    if not os.path.exists(
-        anyscale.project.ANYSCALE_PROJECT_FILE
-    ) and not os.path.exists(project_id_path):
-        # Project id doesn't exist.
-        project_name, project_definition = create_new_proj_def(
-            name, cluster_config_file
-        )
-    else:
+
+    # if config:
+    #     validate_cluster_configuration(config)
+
+    if os.path.exists(project_id_path):
         # Project id exists.
         try:
             project_definition = load_project_or_throw()
@@ -572,7 +623,9 @@ def anyscale_init(
         resp = send_json_request("/api/v2/projects/", {})
         for project in resp["results"]:
             if project["id"] == project_id:
-                if not os.path.exists(anyscale.project.ANYSCALE_AUTOSCALER_FILE):
+                if not os.path.exists(
+                    anyscale.project.ANYSCALE_AUTOSCALER_FILE
+                ) and not os.path.exists(ray_scripts.PROJECT_YAML):
                     # Session yaml file doesn't exist.
                     project_name = get_project_directory_name(project["id"])
                     url = get_endpoint(f"/projects/{project['id']}")
@@ -594,14 +647,63 @@ def anyscale_init(
                     )
         # Project id exists locally but not registered in the db.
         if click.confirm(
-            "The Ansycale project associated with this doesn't "
+            "The Anyscale project associated with this doesn't "
             "seem to exist anymore. Do you want to re-create it?",
             abort=True,
         ):
             os.remove(project_id_path)
-            project_name, project_definition = create_new_proj_def(
-                name, cluster_config_file
+            if os.path.exists(ray_scripts.PROJECT_YAML):
+                with open(ray_scripts.PROJECT_YAML) as f:
+                    proj_yaml = yaml.safe_load(f)
+                if not name and "name" in proj_yaml:
+                    name = proj_yaml["name"]
+                if (
+                    not config
+                    and "cluster" in proj_yaml
+                    and "config" in proj_yaml["cluster"]
+                ):
+                    config = proj_yaml["cluster"]["config"]
+                if name and config and os.path.exists(config):
+                    project_definition = anyscale.project.ProjectDefinition(os.getcwd())
+                    project_definition.config["name"] = slugify(name)
+                    project_definition.config["cluster"]["config"] = config
+                else:
+                    project_name, project_definition = create_new_proj_def(name, config)
+            elif os.path.exists(anyscale.project.ANYSCALE_AUTOSCALER_FILE):
+                project_name, project_definition = create_new_proj_def(
+                    name, project_definition.cluster_yaml()
+                )
+            else:
+                project_name, project_definition = create_new_proj_def(name, config)
+
+    elif os.path.exists(ray_scripts.PROJECT_DIR):
+        # ray-project exists but project-id doesn't exist
+        try:
+            project_definition = load_project_or_throw()
+            if os.path.exists(ray_scripts.PROJECT_YAML):
+                with open(ray_scripts.PROJECT_YAML) as f:
+                    proj_yaml = yaml.safe_load(f)
+                if not name and "name" in proj_yaml:
+                    name = proj_yaml["name"]
+                if (
+                    not config
+                    and "cluster" in proj_yaml
+                    and "config" in proj_yaml["cluster"]
+                ):
+                    config = proj_yaml["cluster"]["config"]
+                if name and config and os.path.exists(config):
+                    project_definition.config["name"] = slugify(name)
+                    project_definition.config["cluster"]["config"] = config
+                else:
+                    raise
+        except Exception:
+            print(
+                "There is an error with the existing ray-project folder, please re-create the project."
             )
+            project_name, project_definition = create_new_proj_def(name, config)
+    else:
+        # Project id doesn't exist and not enough info to create project.
+        project_name, project_definition = create_new_proj_def(name, config)
 
     register_project(project_definition)
 
@@ -685,6 +787,12 @@ def remote_snapshot(
     help="Tag for this snapshot. Multiple tags can be specified by repeating this option.",
     multiple=True,
 )
+@click.option(
+    "--execute-docker-snapshot",
+    is_flag=True,
+    default=False,
+    help="If set, restic is not used",
+)
 def snapshot_create(
     files: List[str],
     description: Optional[str],
@@ -693,6 +801,7 @@ def snapshot_create(
     include_output_files: bool,
     files_only: bool,
     tag: List[str],
+    execute_docker_snapshot: bool,
 ) -> None:
     project_definition = load_project_or_throw()
     project_id = get_project_id(project_definition.root)
@@ -724,6 +833,7 @@ def snapshot_create(
                 additional_files=files,
                 files_only=files_only,
                 tags=tag,
+                run_upload=not execute_docker_snapshot,
             )
         except click.Abort as e:
             raise e
@@ -732,6 +842,8 @@ def snapshot_create(
             # if some files cannot be copied (e.g., due to permissions).
             raise click.ClickException(e)  # type: ignore
 
+    # This is used by the backend to discover the snapshot's UUID
+    print(json.dumps({"snapshot_uuid": snapshot_uuid}))
     # Print success message
     url = get_endpoint(f"/projects/{project_id}")
     print(f"Snapshot {snapshot_uuid} created. View at {url}")
@@ -802,10 +914,7 @@ def session_attach(name: Optional[str], tmux: bool, screen: bool) -> None:
 )
 @click.argument("session-name", required=False)
 @click.option(
-    "--cluster-config-file",
-    "cluster_config_file",
-    help="Cluster to start session with.",
-    default=None,
+    "--config", "config", help="Cluster to start session with.", default=None,
 )
 @click.option(
     "--no-restart",
@@ -837,13 +946,25 @@ def session_attach(name: Optional[str], tmux: bool, screen: bool) -> None:
     type=int,
     help="Override the configured max worker node count for the cluster.",
 )
+@click.option(
+    "--disable-sync",
+    is_flag=True,
+    default=False,
+    help=(
+        "Disables syncing file mounts and project directory. This is "
+        "useful when 'restart-only' is set and file syncing takes a long time."
+    ),
+)
+@click.option("--cloud-id", required=False, help="Id of the cloud to use", default=None)
 def anyscale_up(
     session_name: Optional[str],
-    cluster_config_file: Optional[str],
+    config: Optional[str],
     min_workers: Optional[int],
     max_workers: Optional[int],
     no_restart: bool,
     restart_only: bool,
+    disable_sync: bool,
+    cloud_id: Optional[str],
 ) -> None:
     """Create or update a Ray cluster."""
 
@@ -856,21 +977,35 @@ def anyscale_up(
         )["results"]
         session_name = "session-{0}".format(len(session_list) + 1)
 
-    if not cluster_config_file:
-        cluster_config_file = project_definition.config["cluster"]["config"]
-    cluster_config_file = cast(str, cluster_config_file)
+    if not config:
+        config = project_definition.config["cluster"]["config"]
+    config = cast(str, config)
+    # validate_cluster_configuration(config)
 
-    if not os.path.exists(cluster_config_file):
-        raise ValueError("Project file {} not found".format(cluster_config_file))
-    with open(cluster_config_file) as f:
+    if not os.path.exists(config):
+        raise ValueError("Project file {} not found".format(config))
+    with open(config) as f:
         cluster_config = yaml.safe_load(f)
 
+    enable_cloud = check_is_feature_flag_on(FLAG_KEY_USE_CLOUD)
+    cloud = None
+    if enable_cloud:
+        if cloud_id is None:
+            cloud = get_user_cloud()
+            cloud_id = cloud["id"]
+        else:
+            resp_get_cloud = send_json_request("/api/v2/clouds/{}".format(cloud_id), {})
+            cloud = resp_get_cloud["result"]
+
+        assert cloud is not None, "Failed to get cloud."
+
     resp_out_up = send_json_request(
         "/api/v2/sessions/up",
         {
             "project_id": project_id,
             "name": session_name,
             "cluster_config": {"config": json.dumps(cluster_config)},
+            "cloud_id": cloud_id if enable_cloud else None,
         },
         method="POST",
     )
@@ -879,17 +1014,30 @@ def anyscale_up(
         "/api/v2/sessions/{}".format(resp_out_up["result"]["session_id"]), {}
     )["result"]["head_node_ip"]
 
+    cluster_config = resp_out_up["result"]["cluster_config"]
+
     key_dir = "~/.ssh/session-{session_id}".format(
         session_id=resp_out_up["result"]["session_id"]
     )
-    key_path = write_ssh_key(
-        "ray-autoscaler_4_us-west-2", resp_out_up["result"]["private_key"], key_dir
-    )
+    if not enable_cloud or (cloud is not None and cloud["type"] == "INTERNAL"):
+        # Use this key if cloud is not enabled or if we are using the internal anyscale default cloud.
+        key_name = "ray-autoscaler_4_us-west-2"
+    else:
+        key_name = cluster_config["head_node"]["KeyName"]
+
+    key_path = write_ssh_key(key_name, resp_out_up["result"]["private_key"], key_dir)
 
     cluster_config = resp_out_up["result"]["cluster_config"]
+    cluster_config = update_file_mounts(cluster_config, project_definition)
     cluster_config["auth"].update({"ssh_private_key": key_path})
-    cluster_config["head_node"].update({"KeyName": "ray-autoscaler_4_us-west-2"})
-    cluster_config["worker_nodes"].update({"KeyName": "ray-autoscaler_4_us-west-2"})
+    cluster_config["head_node"].update({"KeyName": key_name})
+    cluster_config["worker_nodes"].update({"KeyName": key_name})
+    docker_in_use = cluster_config.get("docker", {}).get("container_name")
+    if docker_in_use:
+        directory_name = get_project_directory_name(project_id)
+        cluster_config["docker"].setdefault("run_options", []).append(
+            f"-v /home/ubuntu/{directory_name}:/root/{directory_name}"
+        )
 
     resp_out_credentials = send_json_request(
         "/api/v2/sessions/{session_id}/autoscaler_credentials".format(
@@ -901,7 +1049,8 @@ def anyscale_up(
         {"aws_credentials": resp_out_credentials["result"]["credentials"]}
     )
 
-    cluster_config = fillout_defaults(cluster_config)
+    if disable_sync:
+        cluster_config["file_mounts"] = {}
 
     anyscale.util.install_anyscale_hooks(cluster_config)
 
@@ -922,7 +1071,6 @@ def anyscale_up(
                 "--yes",
             ]
             command_lst = [c for c in command if c]
-
             proc = subprocess.Popen(
                 command_lst, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
             )
@@ -935,23 +1083,10 @@ def anyscale_up(
                     print(line, end="")
                     startup_log.append(line)
             startup_log_str = "".join(startup_log)
+            proc.communicate()
 
-            proc = subprocess.Popen(
-                ["anyscale", "list", "ips", session_name, "--json"],
-                stdout=subprocess.PIPE,
-                stderr=subprocess.DEVNULL,
-            )
-            out = proc.communicate()[0].decode()
-            curr_head_ip = json.loads(out[:-1])[0]["head-ip"]
-
-            # Rsync file mounts and project directory after cluster started
-            rsync(
-                config_file.name,
-                None,
-                None,
-                resp_out_up["result"]["cluster_name"],
-                down=False,
-                all_nodes=True,
+            curr_head_ip = get_head_node_ip(
+                config_file.name, resp_out_up["result"]["cluster_name"]
             )
 
             if curr_head_ip != prev_head_ip:
@@ -968,15 +1103,16 @@ def anyscale_up(
                 },
                 "POST",
             )
+            wait_for_session_start(project_id, session_name)
             url = get_endpoint(f"/projects/{project_id}")
             print(f"Session {session_name} started. View at {url}")
-        except Exception:
+        except Exception as e:
             send_json_request(
                 f"/api/v2/sessions/{resp_out_up['result']['session_id']}/stop",
                 {"terminate": True, "workers_only": False, "keep_min_workers": False},
                 method="POST",
             )
-            raise click.ClickException("Session startup failed.")
+            raise click.ClickException("{}\nSession startup failed.".format(e))
 
 
 @click.command(
@@ -1002,7 +1138,7 @@ def anyscale_up(
     default=None,
 )
 @click.option(
-    "--cluster-config-file",
+    "--config",
     help="If set, use this cluster file rather than the default"
     " listed in project.yaml.",
     default=None,
@@ -1032,7 +1168,7 @@ def anyscale_start(
     session_args: str,
     snapshot: Optional[str],
     session_name: Optional[str],
-    cluster_config_file: Optional[str],
+    config: Optional[str],
     min_workers: Optional[int],
     max_workers: Optional[int],
     run: Optional[str],
@@ -1055,8 +1191,8 @@ def anyscale_start(
         session_name = "session-{0}".format(len(session_list) + 1)
 
     # Parse the session arguments.
-    if cluster_config_file:
-        project_definition.config["cluster"]["config"] = cluster_config_file
+    if config:
+        project_definition.config["cluster"]["config"] = config
 
     cluster_params = project_definition.config["cluster"].get("params")
     if cluster_params:
@@ -1164,6 +1300,7 @@ def session_sync(snapshot: Optional[str], name: Optional[str], yes: bool) -> Non
     name="run",
     context_settings=dict(ignore_unknown_options=True,),
     help="Execute a command in a session.",
+    hidden=True,
 )
 @click.argument("command_name", required=False)
 @click.argument("args", nargs=-1, type=click.UNPROCESSED)
@@ -1365,7 +1502,7 @@ def anyscale_autosync(session_name: Optional[str], verbose: bool) -> None:
     directory_name = get_project_directory_name(project_id)
 
     print("Autosync with session {} is starting up...".format(session["name"]))
-    head_ip, key_path = setup_ssh_for_head_node(session["id"])
+    head_ip, key_path, ssh_user = setup_ssh_for_head_node(session["id"])
 
     source = project_definition.root
     target = "~/{}".format(directory_name)
@@ -1388,7 +1525,7 @@ def anyscale_autosync(session_name: Optional[str], verbose: bool) -> None:
                 '"' + " ".join(ssh_command) + '"',
                 "-avz",
                 source,
-                "ubuntu@{}:{}".format(head_ip, target),
+                "{}@{}:{}".format(ssh_user, head_ip, target),
             ]
         )
         subprocess.check_call(command, shell=True)
@@ -1444,7 +1581,7 @@ def anyscale_autosync(session_name: Optional[str], verbose: bool) -> None:
                                 # "--delete-missing-args",
                                 "--files-from={}".format(modified_files.name),
                                 source,
-                                "ubuntu@{}:{}".format(head_ip, target),
+                                "{}@{}:{}".format(ssh_user, head_ip, target),
                             ]
                         )
                         try:
@@ -1462,7 +1599,7 @@ def auth_start() -> None:
     web.run_app(auth_proxy_app)
 
 
-@click.command(name="stop", help="Stop the current session.")
+@click.command(name="down", help="Stop the current session.")
 @click.argument("session-name", required=False, default=None)
 @click.option(
     "--terminate", help="Terminate the session instead of stopping it.", is_flag=True
@@ -1699,10 +1836,41 @@ def session_list(name: Optional[str], all: bool, show_json: bool) -> None:
 
 @click.command(name="pull", help="Pull session")
 @click.argument("session-name", type=str, required=False, default=None)
+@click.option(
+    "--source",
+    "-s",
+    type=str,
+    required=False,
+    default=None,
+    help="Source location to transfer files located on head node of cluster "
+    "from. If source and target are specified, only those files/directories "
+    "will be updated.",
+)
+@click.option(
+    "--target",
+    "-t",
+    type=str,
+    required=False,
+    default=None,
+    help="Local target location to transfer files to. If source and target "
+    "are specified, only those files/directories will be updated.",
+)
+@click.option(
+    "--config",
+    type=str,
+    required=False,
+    default=None,
+    help="Pulls cluster configuration from session this location.",
+)
 @click.confirmation_option(
     prompt="Pulling a session will override the local project directory. Do you want to continue?"
 )
-def anyscale_pull_session(session_name: str) -> None:
+def anyscale_pull_session(
+    session_name: str,
+    source: Optional[str],
+    target: Optional[str],
+    config: Optional[str],
+) -> None:
 
     project_definition = load_project_or_throw()
 
@@ -1717,13 +1885,37 @@ def anyscale_pull_session(session_name: str) -> None:
         with tempfile.NamedTemporaryFile(mode="w") as config_file:
             json.dump(cluster_config, config_file)
             config_file.flush()
-            rsync(
-                config_file.name,
-                source_directory,
-                project_definition.root,
-                None,
-                down=True,
+
+            if source and target:
+                rsync(
+                    config_file.name, source, target, None, down=True,
+                )
+            elif source or target:
+                raise click.ClickException(
+                    "Source and target are not both specified. Please either specify both or neither."
+                )
+            else:
+                rsync(
+                    config_file.name,
+                    source_directory,
+                    project_definition.root,
+                    None,
+                    down=True,
+                )
+
+        if config:
+            session = get_project_session(project_id, session_name)
+            resp = send_json_request(
+                "/api/v2/sessions/{session_id}/cluster_config".format(
+                    session_id=session["id"]
+                ),
+                {},
+                "GET",
             )
+            cluster_config = yaml.safe_load(resp["result"]["config_with_defaults"])
+            with open(config, "w") as f:
+                yaml.dump(cluster_config, f, default_flow_style=False)
+
         print("Pull completed.")
 
     except Exception as e:
@@ -1767,18 +1959,46 @@ def anyscale_pull_snapshot(snapshot_id: str) -> None:
 @click.command(name="push", help="Push current project to session.")
 @click.argument("session-name", type=str, required=False, default=None)
 @click.option(
-    "--snapshot-id",
-    help="Synchronize session with this snapshot rather than taking a new one of the current project.",
+    "--source",
+    "-s",
+    type=str,
+    required=False,
     default=None,
+    help="Source location to transfer files located on head node of cluster "
+    "from. If source and target are specified, only those files/directories "
+    "will be updated.",
 )
 @click.option(
-    "--apply-config",
-    is_flag=True,
-    default=False,
-    help="Take a snapshot and update the cluster if the cluster configuration changes.",
+    "--target",
+    "-t",
+    type=str,
+    required=False,
+    default=None,
+    help="Local target location to transfer files to. If source and target "
+    "are specified, only those files/directories will be updated.",
 )
+@click.option(
+    "--config",
+    type=str,
+    required=False,
+    default=None,
+    help="Updates session with this configuration file.",
+)
+@click.option(
+    "--all-nodes",
+    "-A",
+    is_flag=True,
+    required=False,
+    help="Choose to update to all nodes (workers and head) if source and target are specified.",
+)
+@click.pass_context
 def anyscale_push_session(
-    session_name: str, snapshot_id: Optional[str], apply_config: bool
+    ctx: Any,
+    session_name: str,
+    source: Optional[str],
+    target: Optional[str],
+    config: Optional[str],
+    all_nodes: bool,
 ) -> None:
     project_definition = load_project_or_throw()
     project_id = get_project_id(project_definition.root)
@@ -1790,24 +2010,29 @@ def anyscale_push_session(
     with tempfile.NamedTemporaryFile(mode="w") as config_file:
         json.dump(cluster_config, config_file)
         config_file.flush()
-        rsync(
-            config_file.name, None, None, None, down=False, all_nodes=True,
-        )
 
-    if apply_config:
-        resp = send_json_request(
-            "/api/v2/sessions/{session_id}/take_snapshot".format(
-                session_id=session["id"]
-            ),
-            {"additional_files": [], "files_only": False},
-            method="POST",
-        )
-        snapshot_uuid: str = resp["result"]["id"]
+        if source and target:
+            rsync(
+                config_file.name, source, target, None, down=False, all_nodes=all_nodes,
+            )
+        elif source or target:
+            raise click.ClickException(
+                "Source and target are not both specified. Please either specify both or neither."
+            )
+        else:
+            rsync(
+                config_file.name, None, None, None, down=False, all_nodes=True,
+            )
 
-        send_json_request(
-            "/api/v2/sessions/{session_id}/sync".format(session_id=session["id"]),
-            {"snapshot_id": snapshot_uuid},
-            method="POST",
+    if config:
+        # validate_cluster_configuration(config)
+        print("Updating session with {}".format(config))
+        ctx.invoke(
+            anyscale_up,
+            session_name=session_name,
+            no_restart=False,
+            restart_only=False,
+            disable_sync=True,
         )
 
     url = get_endpoint(f"/projects/{project_id}")
@@ -1890,7 +2115,7 @@ def anyscale_clone(project_name: str) -> None:
 
 
 def clone_files(project_name: str, directory: str, project_id: int) -> None:
-    if check_is_feature_flag_on(FLAG_KEY_USE_SNAPSHOT):
+    if check_is_feature_flag_on(FLAG_KEY_USE_SNAPSHOT, True):
         os.makedirs(os.path.join(directory, "ray-project"))
         with open("{}/ray-project/project-id".format(directory), "w") as f:
             f.write("{}".format(project_id))
@@ -1933,9 +2158,8 @@ def write_ssh_key(key_name: str, key_val: str, file_dir: str = "~/.ssh") -> str:
     file_dir = os.path.expanduser(file_dir)
     os.makedirs(file_dir, exist_ok=True)
     key_path = os.path.join(file_dir, "{}.pem".format(key_name))
-    if not os.path.exists(key_path):
-        with open(os.open(key_path, os.O_WRONLY | os.O_CREAT, 0o600), "w") as f:
-            f.write(key_val)
+    with open(os.open(key_path, os.O_WRONLY | os.O_CREAT, 0o600), "w") as f:
+        f.write(key_val)
     return key_path
 
 
@@ -1945,19 +2169,18 @@ def anyscale_ssh(session_name: str) -> None:
     project_definition = load_project_or_throw()
     project_id = get_project_id(project_definition.root)
     session = get_project_session(project_id, session_name)
-    head_ip, key_path = setup_ssh_for_head_node(session["id"])
+    head_ip, key_path, ssh_user = setup_ssh_for_head_node(session["id"])
 
     cluster_config = get_cluster_config(session_name)
     container_name = cluster_config.get("docker", {}).get("container_name")
 
-    # TODO(ilr) Convert this "ubuntu" to "{config[auth_user][ssh_user]}""
     subprocess.run(
         [
             "ssh",
             "-tt",
             "-i",
             key_path,
-            "ubuntu@{}".format(head_ip),
+            "{}@{}".format(ssh_user, head_ip),
             f"docker exec -it {container_name} sh -c 'which bash && bash || sh'"
             if container_name
             else "",
@@ -1987,20 +2210,48 @@ def get_cluster_config(
         {"aws_credentials": resp["result"]["credentials"]}
     )
     # Get the SSH key from the session.
-    resp = send_json_request("/api/v2/sessions/{}/ssh_key".format(session["id"]), {})
-    key_name = resp["result"]["key_name"]
-    private_key = resp["result"]["private_key"]
-    # Write key to .ssh folder.
-    key_path = write_ssh_key(key_name, private_key)
-    # Store key in autoscaler cluster config.
-    cluster_config["auth"]["ssh_private_key"] = key_path
-    cluster_config["head_node"]["KeyName"] = key_name
-    cluster_config["worker_nodes"]["KeyName"] = key_name
+    existing_key_dir = os.path.expanduser("~/.ssh/session-{}".format(session["id"]))
+    if not os.path.exists(existing_key_dir):
+        resp = send_json_request(
+            "/api/v2/sessions/{}/ssh_key".format(session["id"]), {}
+        )
+        key_name = resp["result"]["key_name"]
+        private_key = resp["result"]["private_key"]
+        # Write key to .ssh folder.
+        key_path = write_ssh_key(key_name, private_key)
+        # Store key in autoscaler cluster config.
+        cluster_config["auth"]["ssh_private_key"] = key_path
+        cluster_config["head_node"]["KeyName"] = key_name
+        cluster_config["worker_nodes"]["KeyName"] = key_name
+    else:
+        # User SSH keys should be something like ~/.ssh/{session_id}/anyscale-user-2_us-west-2_key-4.pem
+        key_files = [
+            f
+            for f in os.listdir(existing_key_dir)
+            if os.path.isfile(os.path.join(existing_key_dir, f))
+        ]
+        assert len(key_files) > 0, "No SSH keys found for session {}".format(
+            session["id"]
+        )
+
+        # Assume there's only one SSH key for this session.
+        key_file = key_files[0]
+        key_file_split = key_file.split(".pem")[0]
+
+        assert len(key_file_split) > 0, "Failed to parse session SSH key name."
+        key_name = key_file_split[0]
+        cluster_config["auth"]["ssh_private_key"] = os.path.join(
+            existing_key_dir, key_file
+        )
+        cluster_config["head_node"]["KeyName"] = key_name
+        cluster_config["worker_nodes"]["KeyName"] = key_name
 
     return cluster_config
 
 
-@cli.command(name="rsync-down", help="Download specific files from cluster.")
+@cli.command(
+    name="rsync-down", help="Download specific files from cluster.", hidden=True
+)
 @click.argument("session-name", required=False, type=str)
 @click.argument("source", required=False, type=str)
 @click.argument("target", required=False, type=str)
@@ -2030,7 +2281,7 @@ def anyscale_rsync_down(
         rsync(config_file.name, source, target, cluster_name, down=True)
 
 
-@cli.command(name="rsync-up", help="Upload specific files to cluster.")
+@cli.command(name="rsync-up", help="Upload specific files to cluster.", hidden=True)
 @click.argument("session-name", required=False, type=str)
 @click.argument("source", required=False, type=str)
 @click.argument("target", required=False, type=str)
@@ -2098,12 +2349,10 @@ def ray_cli(ctx: Any) -> None:
 
         if ctx.invoked_subcommand == "up":
             old_command = "anyscale ray up {}".format(" ".join(args))
-            new_command = "anyscale start --cluster-config-file {}".format(
-                " ".join(args)
-            )
+            new_command = "anyscale start --config {}".format(" ".join(args))
         else:
             old_command = "anyscale ray down {}".format(" ".join(args))
-            new_command = "anyscale stop SESSION_NAME {}".format(" ".join(args[1:]))
+            new_command = "anyscale down SESSION_NAME {}".format(" ".join(args[1:]))
 
         print(
             "\033[91m\nYou called\n  {}\nInstead please call\n  {}\033[00m".format(
@@ -2166,6 +2415,12 @@ def install_autoscaler_shims(ray_cli: Any) -> None:
     default=False,
     help="Rsync all the file mounts before executing the command.",
 )
+@click.option(
+    "--stop",
+    help="Stop session after command finishes executing.",
+    is_flag=True,
+    default=False,
+)
 @click.argument("commands", nargs=-1, type=str)
 def anyscale_exec(
     session_name: str,
@@ -2173,6 +2428,7 @@ def anyscale_exec(
     tmux: bool,
     port_forward: Tuple[int],
     sync: bool,
+    stop: bool,
     commands: List[str],
 ) -> None:
     project_definition = load_project_or_throw()
@@ -2215,12 +2471,12 @@ def anyscale_exec(
         "ANYSCALE_HOST={anyscale_host} anyscale session "
         "upload_command_logs --command-id {session_command_id} {redirect_to_dev_null}; "
         "ANYSCALE_HOST={anyscale_host} anyscale session "
-        "finish_command --command-id {session_command_id} {stop_cmd} {redirect_to_dev_null}".format(
+        "finish_command --command-id {session_command_id} {stop_cmd} {redirect_to_dev_null}; ".format(
             directory_name=directory_name,
             execution_log_name=(execution_log_name(session_command_id)),
             anyscale_host=anyscale.conf.ANYSCALE_HOST,
             session_command_id=session_command_id,
-            stop_cmd="",
+            stop_cmd="--stop" if stop else "",
             shell_command=quote(shell_command),
             redirect_to_dev_null=redirect_to_dev_null,
         )
@@ -2248,6 +2504,7 @@ def anyscale_exec(
             screen=screen,
             tmux=tmux,
             port_forward=[(port, port) for port in list(port_forward)],
+            stop=stop,
         )
 
     if tmux or screen:
