@@ -13,7 +13,13 @@ import click
 import ray.ray_constants
 import requests
 
-from anyscale.conf import SNAPSHOT_REPO, SNAPSHOT_REPO_PASSWORD
+from anyscale.conf import (
+    DOCKER_REGISTRY,
+    DOCKER_SNAPSHOT_BUCKET,
+    REGISTRY_IMAGE,
+    SNAPSHOT_REPO,
+    SNAPSHOT_REPO_PASSWORD,
+)
 from anyscale.feature_flags import FLAG_KEY_USE_SNAPSHOT
 from anyscale.project import get_project_id
 from anyscale.util import (
@@ -29,12 +35,41 @@ from anyscale.util import (
 logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
 logger = logging.getLogger(__file__)
 
+REGISTRY_CONF_PATH = "/etc/docker/registry/config.yml"
+
+REGISTRY_CONF_TEMPLATE = """
+version: 0.1
+log:
+  fields:
+    service: registry
+storage:
+  cache:
+    blobdescriptor: inmemory
+  s3:
+    accesskey: {ACCESS_KEY}
+    secretkey: {SECRET_KEY}
+    sessiontoken: {STS_TOKEN}
+    region: {REGION}
+    bucket: {BUCKET}
+    rootdirectory: {S3_PATH}
+http:
+  addr: :5555
+  headers:
+    X-Content-Type-Options: [nosniff]
+health:
+  storagedriver:
+    enabled: true
+    interval: 10s
+    threshold: 3
+"""
+
 
 class SnapshotRunner:
-    def __init__(self) -> None:
+    def __init__(self, print_snapshot_info: bool = True) -> None:
         self.snapshot_home = os.path.expanduser("~/.anyscale/snapshot")
         self.password = SNAPSHOT_REPO_PASSWORD
         self.snapshot_repo = SNAPSHOT_REPO
+        self.print_snapshot_info = print_snapshot_info
 
         # Get the right restic executable path depending on the OS.
         current_dir = os.path.dirname(os.path.realpath(__file__))
@@ -81,7 +116,6 @@ class SnapshotRunner:
         new_env = os.environ.copy()
         new_env = {**new_env, **creds}
         new_env["RESTIC_PASSWORD"] = self.password
-        print(json.dumps({"snapshot_uuid": snapshot_uuid}))
         proc = subprocess.Popen(
             [
                 self.snapshot_executable,
@@ -103,6 +137,7 @@ class SnapshotRunner:
             stderr=subprocess.PIPE,
         )
         collecting_output = False
+        collected_out = ""
         while True:
             if proc.stdout:
                 line = proc.stdout.readline()
@@ -118,14 +153,16 @@ class SnapshotRunner:
                     timer.start()
                 elif b"scan finished" in line:
                     timer.stop()
-                    print("\n" + line.decode())
+                    if self.print_snapshot_info:
+                        print("\n" + line.decode())
                 elif not collecting_output and line == b"\n":
                     collected_out = ""
                     collecting_output = True
 
                 if collecting_output:
                     collected_out += line.decode()
-        print(collected_out)
+        if self.print_snapshot_info:
+            print(collected_out)
 
     def download(
         self, name: str, target_directory: Union[str, Path], creds: Dict[str, str],
@@ -287,7 +324,9 @@ class SnapshotRunner:
         return parse_restic_snapshots_output(output)
 
 
-snapshot_runner = SnapshotRunner()
+snapshot_runner = SnapshotRunner(
+    print_snapshot_info=check_is_feature_flag_on(FLAG_KEY_USE_SNAPSHOT, True)
+)
 
 
 def parse_restic_snapshots_output(output: bytes) -> Tuple[str, str, List[str]]:
@@ -361,6 +400,7 @@ def create_snapshot(
     additional_files: List[str] = [],
     files_only: bool = False,
     tags: List[str] = [],
+    run_upload: bool = False,
 ) -> str:
     """Create a snapshot of a project.
 
@@ -372,6 +412,7 @@ def create_snapshot(
         additional_files: Additional files to include in the snapshot.
         files_only: Only include additional files in the snapshot.
         tags: Tags for the snapshot.
+        run_upload: Whether to actually upload the restic snapshot.
 
     Raises:
         ValueError: If the current project directory does not match the project
@@ -414,7 +455,8 @@ def create_snapshot(
     )
     snapshot_uuid: str = resp["result"]["id"]
 
-    if check_is_feature_flag_on(FLAG_KEY_USE_SNAPSHOT):
+    if check_is_feature_flag_on(FLAG_KEY_USE_SNAPSHOT, True) and run_upload:
+
         creds = resp["result"]["credentials"]
 
         try:
@@ -520,3 +562,60 @@ def download_snapshot(
     snapshot_runner.download(name, target_directory, credentials)
 
     return target_directory
+
+
+def generate_docker_image_name(project_id: int, snapshot_id: str) -> str:
+    tag = f"{project_id}_{snapshot_id}"
+    return DOCKER_REGISTRY + "/snapshots:" + tag
+
+
+def restore_docker_snapshot_commands(
+    cluster_config: Dict[str, Any],
+    project_id: int,
+    snapshot_id: str,
+    session_id: int,
+    directory_name: str,
+) -> List[str]:
+    new_image_name = generate_docker_image_name(project_id, snapshot_id)
+    cluster_config["docker"]["image"] = new_image_name
+
+    docker_mounts = [
+        "-v {dst}:/root/backup/file_mounts/{dst}:ro"
+        for dst in cluster_config.get("file_mounts", {}).keys()
+    ]
+    docker_mounts.append(
+        f"-v /home/ubuntu/{directory_name}:/root/backup/{directory_name}:ro "
+    )
+
+    restore_commands = [
+        f"mkdir -p {mount}" for mount in cluster_config.get("file_mounts", {}).keys()
+    ]
+
+    restore_commands.append(f"mkdir -p /home/ubuntu/{directory_name}")
+
+    # TODO(ilr) Move away from /root in container.
+    restore_commands.extend(
+        [
+            f"docker run --rm --name temp_backup_{session_id} -d -it -v $HOME/{directory_name}:/root/backup/{directory_name} {new_image_name}-files bash",
+            f"docker exec temp_backup_{session_id} rsync -a /root/save_dir/{directory_name} /root/backup/",
+            f"docker stop temp_backup_{session_id}",
+        ]
+    )
+    return restore_commands
+
+
+def restart_registry_commands(user_aws_credentials: Any, project_id: int,) -> List[str]:
+    registry_config_content = REGISTRY_CONF_TEMPLATE
+    registry_config_content = registry_config_content.format(
+        ACCESS_KEY=user_aws_credentials["AWS_ACCESS_KEY_ID"],
+        SECRET_KEY=user_aws_credentials["AWS_SECRET_ACCESS_KEY"],
+        STS_TOKEN=user_aws_credentials["AWS_SESSION_TOKEN"],
+        REGION="us-west-2",  # TODO: revisit for customer accounts.
+        BUCKET=DOCKER_SNAPSHOT_BUCKET,  # TODO: revisit for customer accounts.
+        S3_PATH=f"docker/{project_id}",
+    )
+
+    return [
+        f"docker stop registry || true && mkdir -p /tmp/docker/registry/ && cat >/tmp/docker/registry/config.yml <<'EOL'{registry_config_content}\nEOL",
+        f"docker run -d -p 5555:5555 -v /tmp/docker/registry/config.yml:{REGISTRY_CONF_PATH} --name registry {REGISTRY_IMAGE}  || docker start registry",
+    ]
