@@ -8,6 +8,7 @@ import threading
 import time
 from typing import Any, cast, Dict, Iterator, List
 import unicodedata
+import webbrowser
 
 import boto3
 import botocore
@@ -230,6 +231,12 @@ def get_available_regions() -> List[str]:
     return [region["RegionName"] for region in client.describe_regions()["Regions"]]
 
 
+def launch_gcp_cloud_setup() -> None:
+    # TODO: Replace this with a proper endpoint
+    endpoint = "/api/v2/clouds/gcp/create/cloud"
+    webbrowser.open(get_endpoint(endpoint))
+
+
 class Timer:
     """
     Code adopted from https://stackoverflow.com/a/39504463/3727678
@@ -268,10 +275,13 @@ class Timer:
         time.sleep(1)
 
 
-def check_is_feature_flag_on(flag_key: str) -> bool:
-    use_snapshot_resp = send_json_request(
-        "/api/v2/userinfo/check_is_feature_flag_on", {"flag_key": flag_key},
-    )
+def check_is_feature_flag_on(flag_key: str, default: bool = False) -> bool:
+    try:
+        use_snapshot_resp = send_json_request(
+            "/api/v2/userinfo/check_is_feature_flag_on", {"flag_key": flag_key},
+        )
+    except Exception:
+        return default
 
     return cast(bool, use_snapshot_resp["result"]["is_on"])
 
@@ -297,7 +307,7 @@ def wait_for_session_start(project_id: int, session_name: str) -> int:
 def install_anyscale_hooks(cluster_config: Dict[str, Any]) -> None:
     "Add anyscale hooks into the cluster configuration."
 
-    cluster_config["setup_commands"].insert(
+    cluster_config.setdefault("setup_commands", []).insert(
         0,
         # The grep line makes sure that the new line
         # is not added again to the ~/.bashrc if it is already there.
